@@ -1,4 +1,4 @@
-# Copyright 2013-2017 The Meson development team
+# Copyright 2013-2019 The Meson development team
 
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -20,21 +20,25 @@ import glob
 import os
 import re
 
-from .. import mesonlib
-from ..mesonlib import version_compare, stringlistify, extract_as_list
+from .. import mesonlib, mlog
+from ..mesonlib import version_compare, stringlistify, extract_as_list, MachineChoice
 from .base import (
     DependencyException, DependencyMethods, ExternalDependency, PkgConfigDependency,
-    strip_system_libdirs, ConfigToolDependency,
+    strip_system_libdirs, ConfigToolDependency, CMakeDependency, HasNativeKwarg
 )
+from .misc import ThreadDependency
+
+from typing import List, Tuple
 
 
-def get_shared_library_suffix(environment, native):
+def get_shared_library_suffix(environment, for_machine: MachineChoice):
     """This is only gauranteed to work for languages that compile to machine
     code, not for languages like C# that use a bytecode and always end in .dll
     """
-    if mesonlib.for_windows(native, environment):
+    m = environment.machines[for_machine]
+    if m.is_windows():
         return '.dll'
-    elif mesonlib.for_darwin(native, environment):
+    elif m.is_darwin():
         return '.dylib'
     return '.so'
 
@@ -45,6 +49,7 @@ class GTestDependency(ExternalDependency):
         self.main = kwargs.get('main', False)
         self.src_dirs = ['/usr/src/gtest/src', '/usr/src/googletest/googletest/src']
         self.detect()
+        self._add_sub_dependency(ThreadDependency, environment, kwargs)
 
     def detect(self):
         gtest_detect = self.clib_compiler.find_library("gtest", self.env, [])
@@ -83,9 +88,6 @@ class GTestDependency(ExternalDependency):
                 return True
         return False
 
-    def need_threads(self):
-        return True
-
     def log_info(self):
         if self.prebuilt:
             return 'prebuilt'
@@ -118,6 +120,7 @@ class GMockDependency(ExternalDependency):
     def __init__(self, environment, kwargs):
         super().__init__('gmock', environment, 'cpp', kwargs)
         self.main = kwargs.get('main', False)
+        self._add_sub_dependency(ThreadDependency, environment, kwargs)
 
         # If we are getting main() from GMock, we definitely
         # want to avoid linking in main() from GTest
@@ -132,10 +135,7 @@ class GMockDependency(ExternalDependency):
         if not gtest_dep.is_found:
             self.is_found = False
             return
-
-        self.compile_args = gtest_dep.compile_args
-        self.link_args = gtest_dep.link_args
-        self.sources = gtest_dep.sources
+        self.ext_deps.append(gtest_dep)
 
         # GMock may be a library or just source.
         # Work with both.
@@ -167,9 +167,6 @@ class GMockDependency(ExternalDependency):
 
         self.is_found = False
 
-    def need_threads(self):
-        return True
-
     def log_info(self):
         if self.prebuilt:
             return 'prebuilt'
@@ -198,35 +195,48 @@ class GMockDependency(ExternalDependency):
         return [DependencyMethods.PKGCONFIG, DependencyMethods.SYSTEM]
 
 
-class LLVMDependency(ConfigToolDependency):
+class LLVMDependencyConfigTool(ConfigToolDependency):
     """
     LLVM uses a special tool, llvm-config, which has arguments for getting
     c args, cxx args, and ldargs as well as version.
     """
-
-    # Ordered list of llvm-config binaries to try. Start with base, then try
-    # newest back to oldest (3.5 is arbitrary), and finally the devel version.
-    # Please note that llvm-config-6.0 is a development snapshot and it should
-    # not be moved to the beginning of the list. The only difference between
-    # llvm-config-8 and llvm-config-devel is that the former is used by
-    # Debian and the latter is used by FreeBSD.
-    tools = [
-        'llvm-config', # base
-        'llvm-config-7',   'llvm-config70',
-        'llvm-config-6.0', 'llvm-config60',
-        'llvm-config-5.0', 'llvm-config50',
-        'llvm-config-4.0', 'llvm-config40',
-        'llvm-config-3.9', 'llvm-config39',
-        'llvm-config-3.8', 'llvm-config38',
-        'llvm-config-3.7', 'llvm-config37',
-        'llvm-config-3.6', 'llvm-config36',
-        'llvm-config-3.5', 'llvm-config35',
-        'llvm-config-8',   'llvm-config-devel', # development snapshot
-    ]
     tool_name = 'llvm-config'
     __cpp_blacklist = {'-DNDEBUG'}
 
     def __init__(self, environment, kwargs):
+        # Already called by `super().__init__`, but need `self.for_machine`
+        # before `super().__init__` is called.
+        HasNativeKwarg.__init__(self, kwargs)
+
+        # Ordered list of llvm-config binaries to try. Start with base, then try
+        # newest back to oldest (3.5 is arbitrary), and finally the devel version.
+        # Please note that llvm-config-6.0 is a development snapshot and it should
+        # not be moved to the beginning of the list.
+        self.tools = [
+            'llvm-config', # base
+            'llvm-config-8',   'llvm-config80',
+            'llvm-config-7',   'llvm-config70',
+            'llvm-config-6.0', 'llvm-config60',
+            'llvm-config-5.0', 'llvm-config50',
+            'llvm-config-4.0', 'llvm-config40',
+            'llvm-config-3.9', 'llvm-config39',
+            'llvm-config-3.8', 'llvm-config38',
+            'llvm-config-3.7', 'llvm-config37',
+            'llvm-config-3.6', 'llvm-config36',
+            'llvm-config-3.5', 'llvm-config35',
+            'llvm-config-9',     # Debian development snapshot
+            'llvm-config-devel', # FreeBSD development snapshot
+        ]
+
+        # Fedora starting with Fedora 30 adds a suffix of the number
+        # of bits in the isa that llvm targets, for example, on x86_64
+        # and aarch64 the name will be llvm-config-64, on x86 and arm
+        # it will be llvm-config-32.
+        if environment.machines[self.for_machine].is_64_bit:
+            self.tools.append('llvm-config-64')
+        else:
+            self.tools.append('llvm-config-32')
+
         # It's necessary for LLVM <= 3.8 to use the C++ linker. For 3.9 and 4.0
         # the C linker works fine if only using the C API.
         super().__init__('LLVM', environment, 'cpp', kwargs)
@@ -250,22 +260,29 @@ class LLVMDependency(ConfigToolDependency):
             self._set_new_link_args(environment)
         else:
             self._set_old_link_args()
-        self.link_args = strip_system_libdirs(environment, self.link_args)
+        self.link_args = strip_system_libdirs(environment, self.for_machine, self.link_args)
         self.link_args = self.__fix_bogus_link_args(self.link_args)
+        self._add_sub_dependency(ThreadDependency, environment, kwargs)
 
-    @staticmethod
-    def __fix_bogus_link_args(args):
+    def __fix_bogus_link_args(self, args):
         """This function attempts to fix bogus link arguments that llvm-config
         generates.
 
         Currently it works around the following:
             - FreeBSD: when statically linking -l/usr/lib/libexecinfo.so will
               be generated, strip the -l in cases like this.
+            - Windows: We may get -LIBPATH:... which is later interpreted as
+              "-L IBPATH:...", if we're using an msvc like compilers convert
+              that to "/LIBPATH", otherwise to "-L ..."
         """
+        cpp = self.env.coredata.compilers[self.for_machine]['cpp']
+
         new_args = []
         for arg in args:
             if arg.startswith('-l') and arg.endswith('.so'):
                 new_args.append(arg.lstrip('-l'))
+            elif arg.startswith('-LIBPATH:'):
+                new_args.extend(cpp.get_linker_search_args(arg.lstrip('-LIBPATH:')))
             else:
                 new_args.append(arg)
         return new_args
@@ -303,7 +320,7 @@ class LLVMDependency(ConfigToolDependency):
             try:
                 self.__check_libfiles(True)
             except DependencyException:
-                lib_ext = get_shared_library_suffix(environment, self.native)
+                lib_ext = get_shared_library_suffix(environment, self.for_machine)
                 libdir = self.get_config_value(['--libdir'], 'link_args')[0]
                 # Sort for reproducability
                 matches = sorted(glob.iglob(os.path.join(libdir, 'libLLVM*{}'.format(lib_ext))))
@@ -389,14 +406,71 @@ class LLVMDependency(ConfigToolDependency):
 
             self.module_details.append(mod + status)
 
-    def need_threads(self):
-        return True
-
     def log_details(self):
         if self.module_details:
             return 'modules: ' + ', '.join(self.module_details)
         return ''
 
+class LLVMDependencyCMake(CMakeDependency):
+    def __init__(self, env, kwargs):
+        self.llvm_modules = stringlistify(extract_as_list(kwargs, 'modules'))
+        self.llvm_opt_modules = stringlistify(extract_as_list(kwargs, 'optional_modules'))
+        super().__init__(name='LLVM', environment=env, language='cpp', kwargs=kwargs)
+
+        # Extract extra include directories and definitions
+        inc_dirs = self.get_cmake_var('PACKAGE_INCLUDE_DIRS')
+        defs = self.get_cmake_var('PACKAGE_DEFINITIONS')
+        temp = ['-I' + x for x in inc_dirs] + defs
+        self.compile_args += [x for x in temp if x not in self.compile_args]
+        self._add_sub_dependency(ThreadDependency, env, kwargs)
+
+    def _main_cmake_file(self) -> str:
+        # Use a custom CMakeLists.txt for LLVM
+        return 'CMakeListsLLVM.txt'
+
+    def _extra_cmake_opts(self) -> List[str]:
+        return ['-DLLVM_MESON_MODULES={}'.format(';'.join(self.llvm_modules + self.llvm_opt_modules))]
+
+    def _map_module_list(self, modules: List[Tuple[str, bool]]) -> List[Tuple[str, bool]]:
+        res = []
+        for mod, required in modules:
+            cm_targets = self.get_cmake_var('MESON_LLVM_TARGETS_{}'.format(mod))
+            if not cm_targets:
+                if required:
+                    raise self._gen_exception('LLVM module {} was not found'.format(mod))
+                else:
+                    mlog.warning('Optional LLVM module', mlog.bold(mod), 'was not found')
+                    continue
+            for i in cm_targets:
+                res += [(i, required)]
+        return res
+
+    def _original_module_name(self, module: str) -> str:
+        orig_name = self.get_cmake_var('MESON_TARGET_TO_LLVM_{}'.format(module))
+        if orig_name:
+            return orig_name[0]
+        return module
+
+class LLVMDependency(ExternalDependency):
+    def __init__(self, env, kwargs):
+        super().__init__('LLVM', env, 'cpp', kwargs)
+
+    @classmethod
+    def _factory(cls, env, kwargs):
+        methods = cls._process_method_kw(kwargs)
+        candidates = []
+
+        if DependencyMethods.CMAKE in methods:
+            candidates.append(functools.partial(LLVMDependencyCMake, env, kwargs))
+
+        if DependencyMethods.CONFIG_TOOL in methods:
+            candidates.append(functools.partial(LLVMDependencyConfigTool, env, kwargs))
+
+        return candidates
+
+    @staticmethod
+    def get_methods():
+        return [DependencyMethods.CMAKE, DependencyMethods.CONFIG_TOOL]
 
 class ValgrindDependency(PkgConfigDependency):
     '''
