@@ -23,14 +23,15 @@ class StaticLinker:
         return mesonlib.is_windows()
 
 
-class VisualStudioLinker(StaticLinker):
+class VisualStudioLikeLinker:
     always_args = ['/NOLOGO']
 
-    def __init__(self, exelist):
+    def __init__(self, exelist, machine):
         self.exelist = exelist
+        self.machine = machine
 
     def get_exelist(self):
-        return self.exelist[:]
+        return self.exelist.copy()
 
     def get_std_link_args(self):
         return []
@@ -39,16 +40,20 @@ class VisualStudioLinker(StaticLinker):
         return []
 
     def get_output_args(self, target):
-        return ['/OUT:' + target]
+        args = []
+        if self.machine:
+            args += ['/MACHINE:' + self.machine]
+        args += ['/OUT:' + target]
+        return args
 
     def get_coverage_link_args(self):
         return []
 
     def get_always_args(self):
-        return VisualStudioLinker.always_args[:]
+        return self.always_args.copy()
 
     def get_linker_always_args(self):
-        return VisualStudioLinker.always_args[:]
+        return self.always_args.copy()
 
     def build_rpath_args(self, build_dir, from_dir, rpath_paths, build_rpath, install_rpath):
         return []
@@ -72,6 +77,16 @@ class VisualStudioLinker(StaticLinker):
         return []
 
 
+class VisualStudioLinker(VisualStudioLikeLinker, StaticLinker):
+
+    """Microsoft's lib static linker."""
+
+
+class IntelVisualStudioLinker(VisualStudioLikeLinker, StaticLinker):
+
+    """Intel's xilib static linker."""
+
+
 class ArLinker(StaticLinker):
 
     def __init__(self, exelist):
